\ SOKOBAN - A MAZE GAME IN FORTH

\ For Sprinter CamelForth 4.01
\ Ported to Sprinter by Alex Hot, 2002.
\ It was made using an ANS sources from GForth package.

40 CONSTANT /MAZE  \ MAXIMAL MAZE LINE

CREATE MAZE  1 CELLS ALLOT /MAZE 25 * ALLOT  \ CURRENT MAZE
VARIABLE MAZES   0 MAZES !  \ ROOT POINTER
VARIABLE SOKO    0 SOKO !   \ PLAYER POSITION
VARIABLE >MAZE   0 >MAZE !  \ CURRENT COMPILED MAZE

\ SCORE INFORMATION

VARIABLE ROCKS     0 ROCKS !  \ NUMBER OF ROCKS LEFT
VARIABLE LEVEL#    0 LEVEL# ! \ CURRENT LEVEL
VARIABLE MOVES     0 MOVES !  \ NUMBER OF MOVES
VARIABLE SCORE     0 SCORE !  \ TOTAL NUMBER OF SCORES

: NEW-MAZE ( N -- ADDR ) \ ADD A NEW LEVEL
    HERE MAZES ROT 1 ?DO  @  LOOP  !
    0 , 0 , HERE >MAZE ! 0 , ;
: COUNT-$ ( ADDR U -- N )  0 ROT ROT
    OVER + SWAP ?DO  I C@ [CHAR] $ = -  LOOP ;
: M: ( "STRING" -- )  \ ADD A LEVEL LINE (TOP FIRST!)
    -1 PARSE TUCK 2DUP COUNT-$ >MAZE @ 1 CELLS - +!
    HERE SWAP MOVE DUP ALLOT
    /MAZE SWAP - HERE OVER BL FILL ALLOT
    >MAZE @ HERE OVER CELL+ - SWAP ! ;

: MAZE-FIELD ( -- ADDR N )
    MAZE DUP CELL+ SWAP @ CHARS ;

: .SCORE ( -- )
    ."  LEVEL: " LEVEL# @ 2 .R ."  SCORE: " SCORE @ 3 .R
    ."  MOVES: " MOVES @ 4 .R ."  ROCKS: " ROCKS @ 2 .R CR CR ;

: .MAZE ( -- )  \ DISPLAY MAZE
    16 9 AT-XY  .SCORE
    CR  MAZE-FIELD OVER + SWAP
    DO  I /MAZE TYPE CR  /MAZE CHARS  +LOOP ;

: FIND-SOKO ( -- N )
    MAZE-FIELD 0
    DO  DUP I CHARS + C@ [CHAR] @ =
	IF  DROP I  UNLOOP  EXIT  THEN
    LOOP  TRUE ABORT" NO PLAYER IN FIELD!" ;

: LEVEL ( N -- FLAG )  \ FINDS LEVEL N
    DUP LEVEL# !
    MAZES  SWAP 0
    ?DO  @  DUP 0= IF  DROP FALSE  UNLOOP  EXIT  THEN  LOOP
    CELL+ DUP @ ROCKS !
    CELL+ DUP @ CELL+ MAZE SWAP CHARS MOVE
    FIND-SOKO SOKO ! TRUE ;

\ NOW THE PLAYING RULES AS REPLACEMENT STRINGS

: 'SOKO ( -- ADDR ) \ GIVES PLAYER'S ADDRESS
    MAZE CELL+ SOKO @ CHARS + ;

: APPLY-RULE? ( ADDR U OFFSET -- FLAG )
    'SOKO 2SWAP
    \ OFFSET SOKO-ADDR ADDR U
    0 DO
	OVER C@ OVER C@ <>
	IF  DROP 2DROP FALSE  UNLOOP  EXIT  THEN
	>R OVER CHARS + R> CHAR+
    LOOP  2DROP DROP  TRUE ;

: APPLY-RULE! ( ADDR U OFFSET -- )
    'SOKO
    2SWAP
    \ OFFSET SOKO-ADDR ADDR U
    0 DO
	COUNT ROT TUCK C! ROT TUCK CHARS + ROT
    LOOP  2DROP DROP ;

: PLAY-RULE ( ADDR1 U1 ADDR2 U2 OFFSET -- FLAG )
    >R 2SWAP R@  APPLY-RULE?
    IF  R> APPLY-RULE! TRUE  ELSE  R> DROP 2DROP FALSE  THEN ;

\ PLAYER MAY MOVE UP, DOWN, LEFT AND RIGHT

: MOVE:  ( OFFSET -- )
    CREATE ,
    DOES>  @ >R  1 MOVES +!
	S" @ "  S"  @"  R@ PLAY-RULE  IF  R> SOKO +!  EXIT  THEN
	S" @."  S"  &"  R@ PLAY-RULE  IF  R> SOKO +!  EXIT  THEN
	S" @$ " S"  @$" R@ PLAY-RULE  IF  R> SOKO +!  EXIT  THEN
	S" @*." S"  &*" R@ PLAY-RULE  IF  R> SOKO +!  EXIT  THEN
	S" @* " S"  &$" R@ PLAY-RULE
	      IF  R> SOKO +!  1 ROCKS +! -1 SCORE +!  EXIT  THEN
	S" @$." S"  @*" R@ PLAY-RULE
	      IF  R> SOKO +! -1 ROCKS +!  1 SCORE +!  EXIT  THEN
	S" &*." S" .&*" R@ PLAY-RULE  IF  R> SOKO +!  EXIT  THEN
	S" &$ " S" .@$" R@ PLAY-RULE  IF  R> SOKO +!  EXIT  THEN
	S" & "  S" .@"  R@ PLAY-RULE  IF  R> SOKO +!  EXIT  THEN
	S" &."  S" .&"  R@ PLAY-RULE  IF  R> SOKO +!  EXIT  THEN
	S" &* " S" .&$" R@ PLAY-RULE
	      IF  R> SOKO +!  1 ROCKS +! -1 SCORE +!  EXIT  THEN
	S" &$." S" .@*" R@ PLAY-RULE
	      IF  R> SOKO +!  1 ROCKS +! -1 SCORE +!  EXIT  THEN
	-1 MOVES +!  R> DROP  ;

1            MOVE: SOKO-RIGHT
-1           MOVE: SOKO-LEFT
/MAZE        MOVE: SOKO-DOWN
/MAZE NEGATE MOVE: SOKO-UP



: PLAY-LOOP ( -- )
    BEGIN
	ROCKS @ 0=
	IF
	    LEVEL# @ 1+ LEVEL  0= IF  EXIT  THEN
	THEN
	.MAZE
	KEY
	CASE
	    [CHAR] Q OF  TRUE		   ENDOF
	    
	    [CHAR] O OF  SOKO-LEFT  FALSE  ENDOF
	    [CHAR] Z OF  SOKO-DOWN  FALSE  ENDOF
	    [CHAR] A OF  SOKO-UP    FALSE  ENDOF
	    [CHAR] P OF  SOKO-RIGHT FALSE  ENDOF

	    [CHAR] o OF  SOKO-LEFT  FALSE  ENDOF
	    [CHAR] z OF  SOKO-DOWN  FALSE  ENDOF
	    [CHAR] a OF  SOKO-UP    FALSE  ENDOF
	    [CHAR] p OF  SOKO-RIGHT FALSE  ENDOF

	    [CHAR] q OF  TRUE              ENDOF
	FALSE SWAP  ENDCASE
    UNTIL ;


TASK: A  TASK: B  TASK: C

CREATE BUF 5 ALLOT


: SOKOBAN   ( n -- ) Start game with any level

	LEVEL IF  PAGE  PLAY-LOOP 
	2 WINDOW PAGE
	26 18 AT-XY
	." G A M E   F I N I S H E D ! " KEY DROP  THEN  PAUSE
;


: TIMES 
 	BEGIN TIME&DATE 2DROP DROP 11 ATTRIB
	67 28 AT-XY  2 .R ." :" 2 .R ." :" 2 .R  PAUSE AGAIN
;

: RULE1  ." You (Soko) are represented by the at-sign @ . " ;

: RULE2  ." Move Soko with A, Z, O or P key. Press Q for exit. " ;

: RULES 31 ATTRIB  
	3 2 AT-XY RULE1
	3 3 AT-XY RULE2 ;
  
' RULES A TASK!
' TIMES C TASK!
' SOKOBAN B TASK!


: GAME ( -- ) \ start game

	11 ATTRIB PAGE

	MULTI   A WAKE  C WAKE  B WAKE

	1 WINDOW 31 ATTRIB
	0 0 80 6 S"  S O K O B A N " OPENPOPUP

 	3 WINDOW 11 ATTRIB
	66 27 10 3 S" " OPENPOPUP

	2 WINDOW 31 ATTRIB
	15 8 48 22 S" " OPENPOPUP

	2 WINDOW 
	." Levels loading..."

	S" levels.txt" INCLUDED

	2 WINDOW PAGE
	." Level? "
	BUF 10 ACCEPT BUF SWAP EVALUATE

	SOKOBAN	
;


: START
	PAGE
	['] GAME CATCH
	?DUP
	IF  PAGE  CR  ." ERROR " . ." ! "
	KEY DROP
	THEN
 	BYE
;


\ ' START IS (COLD)
\ S" soko.exe" SAVE-EXE


